/** 
 *   FIXME ....
 *
 *   Commands definitions for "antenna" position system.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ??? 
 **/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include "command.h"
//#include "encoders.h"
#include "iar.h"
//#include "coordinate.h"
#include "my_stdio.h"


#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

/* A structure which contains information on the commands this program
   can understand. */

typedef struct cm {
	char *name;		/* User printable name of the function. */
	/* Function to call to do the job. */
	int (*func)(int fd_out, char *name);
	char *doc;		/* Documentation for this function.  */
	struct cm *extended;
}COMMAND;

//typedef struct com COMMAND;

/* The names of functions that actually do the manipulation. */
int com_help(int fd_out, char *name);
int com_quit(int fd_out, char *name);
int com_getraw(int fd_out, char *name);
int com_toggle_echo(int fd_out, char *name);

int com_start(int fd_out, char * name);
int com_stop(int fd_out, char * name);
int com_show(int fd_out, char * name);
int com_hide(int fd_out, char * name);
int com_set(int fd_out, char * name);
int com_set_zero(int fd_out, char * name);
int com_set_cw(int fd_out, char * name);
int com_set_ccw(int fd_out, char * name);
int com_test_sclk(int fd_out, char * name);
int com_put_fail_clk(int fd_out, char * name);
int com_put_normal_clk(int fd_out, char * name);

COMMAND commands_set[] = {
	{ "zero", com_set_zero, 
	  "Set zero for current position on encoder \"dec\" or \"ra\"" },
	{ "cw", com_set_cw, 
	  "Set clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ "ccw", com_set_ccw, 
	  "Set counter clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ (char *)NULL, (int) NULL, (char *)NULL }
};

COMMAND commands_main[] = {
	{ "set", com_set, 
	  "Set some encoders attributes (see help)", commands_set},
	{ "start", com_start, "Start acquisition", NULL},
	{ "stop", com_stop, "Stop acquisition", NULL },
	{ "show", com_show, "Show positon continuously"},
	{ "hide", com_hide, "Not show positon continuously"},
	{ "getenc", com_getraw, "Get raw position of encoders" },
	{ "testsclk", com_test_sclk, "Test if SCLK is running."},
	{ "failclk", com_put_fail_clk, 
	  "Put system in fail condition when SCLK not running"},
	{ "normalclk", com_put_normal_clk,
	  "Put system in normal condition when SCLK running again"},
	{ "e", com_toggle_echo, "Toggle remote echo (for Win telnet clients)"},
	{ "help", com_help, "Display this text"},
	{ "?", com_help, "Synonym for `help'"},
	{ "quit", com_quit, "Quit ..."},
//	{ "q", com_quit, "Quit ..."},
	{ (char *)NULL, (int) NULL, (char *)NULL}
};


/* When non-zero, this means the user is done using this program. */
static int done = 0;
static int remote_echo = 0;
static int start_survey = 0;
static int track_pos = 0;
static int fd_save = -1;
static char *file_save;

#ifdef USE_READLINE

struct list_index{
	int index;
	int len;
	struct list_index *next;
};

/* Generator function for command completion.  STATE lets us
   know whether to start from scratch; without any state
   (i.e. STATE == 0), then we start at the top of the list. */
char *command_generator_real (char *text, int state, 
			      struct list_index *l, COMMAND *cm)
{ 
	char *name;

	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if (!state) {
		l->index = 0;
		l->len = strlen (text);
	}

	/* Return the next name which partially matches from the
	   command list. */
	while ((name = cm[l->index].name)) {
		l->index++;
		if (strncmp (name, text, l->len) == 0 ){
			return (strdup(name));
		}
	}

	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}



char *command_generator (char *text, int state)
{ 
	static struct list_index l={0,0,NULL};

	return command_generator_real(text, state, &l, commands_main);
}


char **fileman_completion (const char *text, int start, int end)
{
	char **matches;
	char *all_text;
	matches = (char **)NULL;

	all_text = rl_copy_text(0,end );

	/* If this word is at the start of the line, then it is a
	   command to complete.  Otherwise it is the name of a file in
	   the current directory.  */
	//if (start == 0)

	matches = rl_completion_matches (all_text, 
					 (rl_compentry_func_t *) 
					 command_generator);
	free(all_text);
	  
	return (matches);
}


/* Tell the GNU Readline library how to complete.  We want to try to
   complete on command names if this is the first word in the line, or
   on filenames if not. */
void initialize_readline ()
{
	/* Allow conditional parsing of the ~/.inputrc file. */
	/* FIXME */
	rl_readline_name = "antenad";
	
	rl_filename_completion_desired = 0;

	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = fileman_completion;
	//rl_attempted_completion_function = NULL;;
	rl_completion_entry_function = NULL;
}

#endif



/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
COMMAND *find_command (COMMAND *cm, char *name)
{
	int i;

	for (i = 0; cm[i].name; i++)
		if (strcmp (name, cm[i].name) == 0)
			return (&cm[i]);
	
	return ((COMMAND *)NULL);
}


/* Execute a command line. */
int execute_line (int fd_out, COMMAND *cm, char *line)
{
	int i;
	COMMAND *command;
	char *word;
	
	/* Isolate the command word. */
	i = 0;
	while (line[i] && whitespace (line[i]))	i++;
	word = line + i;
	
	while (line[i] && !whitespace (line[i])) i++;

	if (line[i])
		line[i++] = '\0';

	command = find_command (cm, word);

	if (!command) {
		fd_printf (fd_out, 
			   "# %s: No such command for ""FIXME"".\n", word);
		return (-1);
	}

	/* Get argument to command, if any. */
	while (whitespace (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (fd_out, word));
}



/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (int fd_out, char *arg)
{
	done = 1;
	return (0);
}


int com_toggle_echo (int fd_out, char *arg)
{
	remote_echo ^= 1;
	return (0);
}


/* Print out help for ARG, or for all of the commands if ARG is not
   present. */

int com_help_real (int fd_out, char *arg, COMMAND *cm)
{
	int i, j;
	int printed = 0;
	int cmp;
	char *sub_arg = "";
	COMMAND *commands = cm;
	
	for (i = 0; commands[i].name; i++) {
		cmp = strcmp (arg, commands[i].name);
		/* FIXME: set se parece a otros comando y cmp retoran valores positivos */ 
		//printf("cmp: %d\n", cmp);
		if (!*arg || ( cmp == 0)) {
			fd_printf (fd_out, "# %-13s%s.\n", commands[i].name, 
				commands[i].doc);
			printed++;
		}
		/* FIXME cont: originalmente (cmp >= 0) */
		if ( commands[i].extended != NULL && (cmp == 0)){
			j = -1;
			if (cmp > 0){
				j = strlen(commands[i].name) + 1;
				while (whitespace (arg[j])) j++;
				sub_arg = arg + j;
			}
			if ( (j + 1) < strlen(arg) ){
				if ( sub_arg == "" )
					fd_printf (fd_out,
						   "# Possible %s options ar:\n", arg);
				com_help_real(fd_out, sub_arg, commands->extended);
				return 0;
			}
		}
	}

	if (!printed) {
		fd_printf (fd_out, 
			   "# No commands match `%s'. Possibilties are:\n", 
			   arg);

		fd_printf (fd_out, "# ");
		for (i = 0; commands[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				fd_printf (fd_out, "\n# ");
			}
			
			fd_printf (fd_out, "%-13s", commands[i].name);
			printed++;
		}

		if (printed)
			fd_printf (fd_out, "\n");
	}
	return (0);
}


/* help front end */

int com_help (int fd_out, char *arg)
{
	return com_help_real(fd_out, arg, commands_main);
}


int open_save_file(char *buff)
{
	/* file_save */
	char data_dir[] = {"."};
	char *np = buff;
	char *endp;
	int alfa_norm, delta_norm;
	int grid = 5;
  	time_t now;
	struct tm *t;

	file_save = (char *) malloc(160);
	alfa_norm =grid_norm(strtod(np, &endp ), grid);
	np = endp + 1;
	delta_norm =grid_norm(strtod(np, &endp ), grid);
	/* Create date stamp and open data file */
	time(&now);
	t = localtime(&now);
	sprintf(file_save, "%s/%04d%02d%02d_%03d_%03d.dat", 
		data_dir, 
		t->tm_year + 1900, t->tm_mon+1, t->tm_mday,
		alfa_norm, delta_norm );
	fd_save = open(file_save, O_WRONLY|O_CREAT|O_APPEND/*|O_SYNC*/,
		       S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd_save >= 0) {
		/* print header */
		fd_printf(fd_save,"# Antenna 2 position survey\n");
		fd_printf(fd_save,"# %04d/%02d/%02d %02d:%02d:%02d\n",
			  t->tm_year + 1900, t->tm_mon+1, t->tm_mday,
			  t->tm_hour, t->tm_min, t->tm_sec);
		fd_printf(fd_save,"# pos: %03d %03d\n#\n", 
			  alfa_norm, delta_norm );
	}

	return fd_save;
}


int manager_track(int fd_out, int fd_track)
{
	char buff[PIPE_BUFFER_LEN];
	int l;
	
	l = read(fd_track, buff, PIPE_BUFFER_LEN);
	//fprintf(stderr,"Read track: [%s]\n",buff);
	if ( track_pos ) {
		if (buff[0] == POSITION_PREFIX)
			write(fd_out, buff+2, l-2);
		else 
			write(fd_out, buff, l);
	}
	else {
		if (buff[0] != POSITION_PREFIX)
			write(fd_out, buff, l);
	}
	if( start_survey && buff[0] == POSITION_PREFIX) {
		if (fd_save < 0) {
			/* init file */
			fd_save = open_save_file(buff+2);
			if (fd_save < 0) {
				perror("open save file:");
				fd_printf(fd_out,"# Error opening: %s\n",
					  file_save);
				start_survey = 0;
			}
			else {
				fd_printf(fd_out,"# Save data in: %s\n",
					  file_save);
				fd_printf(fd_track,"m:# Saving...\n");
			}

		}
		if ( !(fd_save < 0)) 
			write(fd_save, buff+2, l-2);
	}

	return 0;
}


int command_main(int fd_in, int fd_out, int fd_track)
{
	char *line = NULL,
		*s;
	fd_set rfds;
	struct timeval tv;
	int retval;

#ifndef USE_READLINE
	char c = 0;
	int l;
#endif
	done = 0;
#ifdef USE_READLINE
	initialize_readline (); /* Bind our completer. */
#else
	line = (char *) malloc(COMMAND_LINE_LEGEND*sizeof(char));
#endif
	/* Loop reading and executing lines until the user quits. */

	write(fd_track,"s\n",2);

	while( !done ) {
#ifdef USE_READLINE
		line = readline ("TEST: ");
#else
		//l = read(fd_in, line, COMMAND_LINE_LEGEND);
		l = 0;
		do {

			FD_ZERO(&rfds);
			FD_SET(fd_in, &rfds);
			if(fd_track) FD_SET(fd_track, &rfds);
			tv.tv_sec = 10;
			tv.tv_usec = 0;
			retval = select(50, &rfds, NULL, NULL, &tv);
			c = 0;
			if (retval) {
				if (FD_ISSET(fd_in, &rfds)) {
				/* Win telnet clients send char by char ... */
					read(fd_in, &c, 1); 
					*(line+l) = c;
					l++;
					if ( remote_echo ) 
						fd_printf(fd_out,"%c", c);
				}
				if (FD_ISSET(fd_track, &rfds)) {
					manager_track(fd_out, fd_track);
				}
			}
		} while(c != '\n' && l < COMMAND_LINE_LEGEND);
		if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
		else *(line+l-1) = 0;
#endif

		if (!line)
			break;
		/* Remove leading and trailing whitespace from the
		   line.  Then, if there is anything left, add it to
		   the history list and execute it. */
		s = stripwhite (line);
		if (*s){
#ifdef USE_READLINE
			add_history (s);
#endif
			execute_line (fd_out, commands_main, s);
		}
		
	}
	write(fd_track,"q\n",2);		

	free (line);
	/* fd_printf (fd_out, "# connection closed\n"); */

	return 0;
}


int com_set(int fd_out, char * name)
{
	if ( *name == '\0') {
		fd_printf (fd_out, "# set: missing argument.\n");
		return (-1);
	}

	return execute_line(fd_out, commands_set, name);
}


int com_start(int fd_out, char * name)
{

	if (start_survey) {
		fd_printf(fd_out,"# Error: now in start mode.\n");
	}
	start_survey = 1;

	return 0;
}


int com_stop(int fd_out, char * name)
{
	if (start_survey) {
		fd_printf(fd_out,"# closing: %s\n", file_save);
		close(fd_save);
		free(file_save);
		fd_save = -1;
		start_survey = 0;
	}

	return 0;
}


int com_show(int fd_out, char * name)
{
	track_pos = 1;
	return 0;
}

int com_hide(int fd_out, char * name)
{
	track_pos = 0;
	return 0;
}
